from .pipes import (
    dataframe_external_funcs,
    dataframe_instance_methods,
    general_external_funcs,
    general_instance_methods,
)

__all__ = [
    "general_external_funcs",
    "general_instance_methods",
    "dataframe_external_funcs",
    "dataframe_instance_methods",
]
